
// Copyright (c) 2008 Dean Michael Berris
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_SELECT_20080417
#define BOOST_SELECT_20080417

#include <functional>
#include <utility>
#include <boost/type_traits/remove_reference.hpp>

namespace boost {

	template <const unsigned int N>
	struct select;

	namespace detail {

		//////////////////////////////////////////////////////////////////
		// Forward declaration of the metafunction which chooses
		// which type should be returned given two types and
		// the position indicator.
		//
		template <class T1, class T2, const unsigned int N>
		struct result;
		
		// Specialization for the first type
		template <class T1, class T2>
		struct result<T1, T2, 0u> {
			typedef typename boost::remove_reference<T1>::type type;
		};

		// Specialization for the second type
		template <class T1, class T2>
		struct result<T1, T2, 1u> {
			typedef typename boost::remove_reference<T2>::type type;
		};
		//
		//////////////////////////////////////////////////////////////////

		//////////////////////////////////////////////////////////////////
		// Forward declaration of the wrapper type which is convertible
		// to the correct type given the pair<T1, T2> and a positional
		// argument N.
		//
		template <class T1, class T2, const unsigned int N>
		struct get_impl;

		// Specialization for the first type
		template <class T1, class T2>
		struct get_impl<T1, T2, 0u> {

		private:
			typedef typename result<T1, T2, 0u>::type result_type;
			std::pair<T1, T2> const & _pair;

		public:
			explicit get_impl(std::pair<T1, T2> const & pair)
				: _pair(pair)
			{ };

			operator result_type const &() {
				return _pair.first;
			};

		};

		// Specialization for the second type
		template <class T1, class T2>
		struct get_impl<T1, T2, 1u> {
		private:
			typedef typename result<T1, T2, 1u>::type result_type;
			std::pair<T1, T2> const & _pair;

		public:
			explicit get_impl(std::pair<T1, T2> const & pair)
				: _pair(pair)
			{ };

			operator result_type const & () {
				return _pair.second;
			};

		
		};

		// Factory function, where T1 and T2 are deduced
		// and N is specified upon use.
		//
		template <const unsigned int N, class T1, class T2>
		get_impl<T1, T2, N> get(std::pair<T1, T2> const & input) {
			return get_impl<T1, T2, N>(input);
		};
		//
		//////////////////////////////////////////////////////////////////

	}; // namespace detail
	
	//////////////////////////////////////////////////////////////////
	// The polymorphic select template function object which takes 
	// a reference to a const pair as the sole argument to the 
	// function operator that returns a reference to the first 
	// (select<0>) or second (select<1>) element of the pair.
	//
	template <const unsigned int N>
	struct select {
		template <class T1, class T2>
		typename detail::result<T1, T2, N>::type const & operator()(std::pair<T1, T2> const & input) const {
			return detail::get<N>(input);
		};
	};
	//
	//////////////////////////////////////////////////////////////////

	typedef select<0> first; // convenience selector

	typedef select<1> second; // convenience selector

}; // namespace boost

#endif